/* $Id: checkfmt.c,v 1.3 2000/10/10 22:10:44 dps Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif /* HAVE_CTYPE_H */
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif /* HAVE_STDIO_H */
#include "libdetail/detail.h"
#include "prog.h"

#define LESS 1
#define EQUAL 2
#define MORE 4
#define DEFAULT_MASK EQUAL

static int mask_check(int ok_mask, int n, int m)
{
    int mask;

    if (ok_mask==0)
	ok_mask=DEFAULT_MASK;

    if (n<m)
	mask=LESS;
    else if (n==m)
	mask=EQUAL;
    else
	mask=MORE;

    return ((ok_mask & mask)!=0) ? 1 : 0;
}
		

/* Enforce format match with printf style spec string */
int check_fmt(FILE *in, const unsigned char *fmt)
{
    int c, digits, l, len;
    int ok_mask;		/* Whether <, =, > is ok */

    while(*fmt!='\0')
    {
	switch(*fmt)
	{
	case '!':			/* Optional white space */
	    if (absorb_ws(in)==-1)
		return 0;
	    break;
	case ' ':			/* Forced white space */
	    if (absorb_ws(in)==0)
		return 0;
	    break;
	case '%':
	    fmt++;
	    l=0;
	    while(isdigit(*fmt))
	    {
		l=l*10+(*fmt)-'0';
		fmt++;
	    }	    
	    ok_mask=0;
	fmt_switch:
	    switch(*fmt)
	    {
	    case '<':
		ok_mask |= LESS;
		fmt++;
		goto fmt_switch;
		/* NOT REACHED */

	    case '=':
		ok_mask |= EQUAL;
		fmt++;
		goto fmt_switch;
		/* NOT REACHED */

	    case '>':
		ok_mask |= MORE;
		fmt++;
		goto fmt_switch;
		/* NOT REACHED */

	    case '\0':
		fputs("Invalid spec passed to check_fmt", stderr);
		return 1;		/* Matched */
		/* NOT REACHED */

	    case 'x':
		(void) readhex(in, &digits);
		if (digits==0)
		    return 0;		/* Zero digits means no match */
		if (l==0)
		    break;		/* If length not specifed the OK */
		if (!mask_check(ok_mask, digits, l))
		    return 0;		/* Others wrong number of digits=no match */
		break;

	    case 'd':
		(void) readdec(in, &digits);
		if (digits==0)
		    return 0;		/* Zero digits means no match */
		if (l==0)
		    break;		/* If length not specifed then OK */
		if (!mask_check(ok_mask, digits, l))
		    return 0;		/* Others wrong number of digits=no match */
		break;

	    case 's':
		if (l==0)
		    ok_mask=LESS | EQUAL | MORE;
		if ((len=skip_string(in, MAX_STRING))==-1)
		    return 0;
		if (!mask_check(ok_mask, len, l))
		    return 0;
		break;

	    default:			/* Otherwise % is quote character */
		if ((c=fgetc(in))!=(*fmt))
		{
		    ungetc(c, in);	/* Push back c */
		    return 0;		/* Non matched */
		}
		break;
	    }
	    break;


	default:
	    if ((c=fgetc(in))!=(*fmt))
	    {
		ungetc(c, in);		/* Push back c */
		return 0;		/* Non matched */
	    }
	    break;
	}
#ifdef DEBUG
	fprintf(stderr, "Passed fmt=%s\n", fmt);
#endif /* DEBUG */
	fmt++;
    }
    return 1;				/* Input matched */
}
